@@ -32,7 +32,6 @@ class KakiRequestHandler
 
 		data = {}
 		begin
-
 			# template file
 			tmpl_file = "#{command}.html"
 
@@ -47,17 +46,19 @@ class KakiRequestHandler
 			when 'save' then
 				save(request_header, request_parameters, data)
 			when 'change_passwd' then
+				tmpl_file = "save.html"
 				change_passwd(request_header, request_parameters, data)
-				tmpl_file = "save.html"
 			when 'snapshot' then
-				snapshot(request_header, request_parameters, data)
 				tmpl_file = "save.html"
+				snapshot(request_header, request_parameters, data)
 			when 'index' then
 				index(request_header, request_parameters, data)
 			when 'recent' then
 				recent(request_header, request_parameters, data)
 			when 'admin' then
 				admin(request_header, request_parameters, data)
+			when 'delete' then
+				delete(request_header, request_parameters, data)
 			else 
 				# default command is 'view'
 				command = 'view'
@@ -65,8 +66,6 @@ class KakiRequestHandler
 				view(request_header, request_parameters, data)
 			end
 
-			# common data 
-			data.update(site_common_data())
 
 			# side menu
 			request_lang_list = self.reuqest_language_list(request_header)
@@ -89,7 +88,7 @@ class KakiRequestHandler
 					command = 'edit'
 					retry
 				else
-					data[:exception] = exception2html($!) if $! 
+					data[:exception] = $!.to_html if $! 
 				end
 			end
 		rescue 
@@ -103,28 +102,34 @@ class KakiRequestHandler
 			when "NoSuchPage" then
 				if command == 'view' then
 					command = 'edit'
+					request_parameters['p'] = ['FrontPage']
 					retry
 				else
-					data[:exception] = exception2html($!) if $! 
+					data[:exception] = $!.to_html if $! 
 				end
 			when "PageAuthorizationFail" then
-				if command == 'save' then
+				case command
+				when 'save' then
 					command = 'preview'
 					retry
-				elsif command == 'snapshot' then
+				when 'snapshot', 'change_passwd' then
 					command = 'edit'
 					retry
 				else
-					data[:exception] = exception2html($!) if $! 
+					data[:exception] = $!.to_html if $! 
 				end
 			else
-				data[:exception] = exception2html($!) if $! 
+				data[:exception] = $!.to_html if $! 
 			end
 		else 
-			data[:exception] = exception2html($!) if $! 
+			data[:exception] = $!.to_html if $! 
 		end
 
+		# common data 
+		data.update(site_common_data())
+
 		# template file path
+		request_lang_list = [] unless request_lang_list
 		tmpl_path, lang = template_path(request_lang_list + ['en'], tmpl_file)
 		tmpl =  File.readlines(tmpl_path).to_s
 
@@ -181,18 +186,28 @@ protected
 			data[:content] = @wiki2html.output 
 			@page_data_source.page_property(page_name, page_lang, page_version){ |property|
 				passwd = property['passwd']
-				data[:is_freezed] = true if (passwd and passwd.length > 0)
+				data[:is_frozen] = self.site_property['is_frozen']
+				data[:is_frozen] = true if (passwd and passwd.length > 0)
 				data[:counter] = { :count => property['count_int']}
 				data[:source_time] = property['source_time']
-				data[:page_title] = property['page_title']
-				data[:modified_time] = time2hash(property['source_time'])
+				data[:page_title] = CGI.escapeHTML(property['page_title'])
+				data[:modified_time] = property['source_time'].to_data_hash
+				if request_header['referer'] then
+					url = request_header['referer'].to_s 
+					referers = property['referers'].to_s.split(", ")
+					referers.delete(url)
+					referers.unshift(url)
+					referers = referers[0,10]
+					property['referers'] = referers.join(", ")
+				end
 				property['count_int'] += 1
 			}
 			
-			data[:page_id] = page_id
-			data[:page_name] = page_name
-			data[:page_lang] = page_lang
-			data[:page_name_and_lang] = @page_data_source.make_page_id(page_name, page_lang)
+			data[:page_id_uri] = URI.escape(page_id)
+			data[:page_id] = CGI.escapeHTML(page_id)
+			data[:page_name] = CGI.escapeHTML(page_name)
+			data[:page_lang] = CGI.escapeHTML(page_lang)
+			data[:page_name_and_lang] = URI.escape(@page_data_source.make_page_id(page_name, page_lang))
 			if page_version then
 				data[:page_version] = { :version_int => page_version.to_i } 
 			end
@@ -227,28 +242,28 @@ protected
 		source_time_str = '0'
 		if @page_data_source.exist?(page_name, page_lang) then
 			content_source = @page_data_source.load_page_source(page_name, page_lang);
-			content_source = CGI.escapeHTML(content_source);
 			@page_data_source.page_property(page_name, page_lang) { |property|
 				page_title = property['page_title']
 				source_time_str = property['source_time'].to_i.to_s.reverse
 				passwd = property['passwd']
-				data[:is_freezed] = true if (passwd and passwd.length > 0)
+				data[:is_frozen] = self.site_property['is_frozen']
+				data[:is_frozen] = true if (passwd and passwd.length > 0)
 			}
 		else
 			original_language = request_parameters['original_language'].to_s.untaint
 			if original_language.length > 0 and @page_data_source.exist?(page_name, original_language) then
 				content_source = @page_data_source.load_page_source(page_name, original_language);
-				content_source = CGI.escapeHTML(content_source);
 			end
 		end
 
-		data[:page_id] = page_id
-		data[:page_name] = page_name
-		data[:page_lang] = page_lang
-		data[:page_title] = page_title
-		data[:is_not_freezed] = (not data[:is_freezed])
+		data[:page_id_uri] = URI.escape(page_id)
+		data[:page_id] = CGI.escapeHTML(page_id)
+		data[:page_name] = CGI.escapeHTML(page_name)
+		data[:page_lang] = CGI.escapeHTML(page_lang)
+		data[:page_title] = CGI.escapeHTML(page_title)
+		data[:is_not_frozen] = (not data[:is_frozen])
 
-		data[:content_source] = content_source
+		data[:content_source] = CGI.escapeHTML(content_source)
 		salt = [[rand(4096)].pack("v")].pack("m").tr("+", ".")
 		data[:edit_key] = source_time_str.crypt(salt)
 	end
@@ -264,22 +279,23 @@ protected
 		page_id = @page_data_source.make_page_id(page_name, page_lang)
 		page_title = request_page_title(request_parameters)
 
-		is_freezed = false
+		is_frozen = self.site_property['is_frozen']
 		if @page_data_source.exist?(page_name, page_lang) then
 			@page_data_source.page_property(page_name, page_lang, page_version){ |property|
 				if property['passwd'] and property['passwd'].length > 0 then
-					is_freezed = true
+					is_frozen = true
 				end
 			}
 		end
 
-		data[:page_id] = page_id
-		data[:page_name] = page_name
-		data[:page_lang] = page_lang
-		data[:page_title] = page_title
+		data[:page_id_uri] = URI.escape(page_id)
+		data[:page_id] = CGI.escapeHTML(page_id)
+		data[:page_name] = CGI.escapeHTML(page_name)
+		data[:page_lang] = CGI.escapeHTML(page_lang)
+		data[:page_title] = CGI.escapeHTML(page_title)
 
-		data[:is_freezed] = is_freezed
-		data[:content_source] = content_source
+		data[:is_frozen] = is_frozen
+		data[:content_source] = CGI.escapeHTML(content_source)
 		data[:edit_key] = edit_key
 
 		wiki_doc =  Wiki::PageParser.new().parse(content_source);
@@ -309,10 +325,7 @@ protected
 			property = @page_data_source.page_property(page_name, page_lang)
 		end
 
-		if self.page_login_ok?(property, passwd) or
-			(admin_passwd_exist?(@page_data_source.site_property()) and 
-			admin_login_ok?(@page_data_source.site_property(), passwd)) then
-
+		if self.page_login_ok?(property, passwd) then
 			# check editing conflict 
 			source_time_str = property['source_time'].to_i.to_s.reverse if property 
 			if edit_key != source_time_str.crypt(edit_key) then
@@ -328,10 +341,12 @@ protected
 		else
 			data[:page_save_auth_fail] = true
 		end
-		data[:page_id] = page_id
-		data[:page_name] = page_name
-		data[:page_lang] = page_lang
-		data[:page_title] = property['page_title']
+
+		data[:page_id_uri] = URI.escape(page_id)
+		data[:page_id] = CGI.escapeHTML(page_id)
+		data[:page_name] = CGI.escapeHTML(page_name)
+		data[:page_lang] = CGI.escapeHTML(page_lang)
+		data[:page_title] = CGI.escapeHTML(property['page_title'])
 
 		if data[:page_save_auth_fail] then
 			raise "PageAuthorizationFail"
@@ -347,34 +362,37 @@ protected
 
 		@page_data_source.page_property(page_name, page_lang){ |property|
 			if self.page_login_ok?(property, passwd) then
-					# if non passwd or authorized
-					if new_passwd1.length > 0 then
-						if new_passwd1 == new_passwd2 then
-							if property['passwd'].to_s.length > 0 then
-								data[:passwd_change_ok] = { :page_id => page_id, :page_name => page_name }
-							else
-								data[:freeze_ok] = { :page_id => page_id, :page_name => page_name }
-							end
-
-							salt = [[rand(4096)].pack("v")].pack("m").tr("+", ".")
-							property['passwd'] = new_passwd1.crypt(salt)
+				# if non passwd or authorized
+				if new_passwd1.length > 0 then
+					if new_passwd1 == new_passwd2 then
+						if property['passwd'].to_s.length > 0 then
+							data[:passwd_change_ok] = true
 						else
-							data[:new_passwd_miss_match] = true
+							data[:freeze_ok] = { :page_id => CGI.escapeHTML(page_id) }
 						end
+
+						salt = [[rand(4096)].pack("v")].pack("m").tr("+", ".")
+						property['passwd'] = new_passwd1.crypt(salt)
 					else
-						property['passwd'] = nil
-						data[:thaw_ok] = { :page_id => page_id, :page_name => page_name }
+						data[:new_passwd_miss_match] = true
 					end
+				else
+					property['passwd'] = nil
+					data[:thaw_ok] = { :page_id => CGI.escapeHTML(page_id) }
+				end
 			else
-				data[:auth_fail_msg] = true
-				raise "PageAuthorizationFail"
+				data[:auth_fail] = true
 			end
-			data[:page_id] = page_id
-			data[:page_name] = page_name
-			data[:page_lang] = page_lang
-			data[:page_title] = property['page_title']
+			data[:page_id_uri] = URI.escape(page_id)
+			data[:page_id] = CGI.escapeHTML(page_id)
+			data[:page_name] = CGI.escapeHTML(page_name)
+			data[:page_lang] = CGI.escapeHTML(page_lang)
+			data[:page_title] = CGI.escapeHTML(property['page_title'])
 		}
 
+		if data[:auth_fail] then
+			raise "PageAuthorizationFail"
+		end
 	end
 
 	def snapshot(request_header, request_parameters, data)
@@ -383,7 +401,7 @@ protected
 		page_name, page_lang, = @page_data_source.parse_page_id(page_id)
 
 		@page_data_source.page_property(page_name, page_lang){ |property|
-			if self.page_login_ok?(page_id, passwd) then
+			if self.page_login_ok?(property, passwd) then
 				version_str = sprintf("%08d", property['version_int'])
 				source = @page_data_source.load_page_source(page_name, page_lang)
 				@page_data_source.save_page_source(page_name, page_lang, version_str, source)
@@ -391,19 +409,24 @@ protected
 					property.each(){ |key, value|
 						snap_prop[key] = value
 					}
+					snap_prop['passwd'] = nil
 				}
 				property['version_int'] += 1
-				data[:snapshot_ok_msg] = { :page_id => page_id, :page_name => page_name }
-				data[:page_id] = page_id
-				data[:page_name] = page_name
-				data[:page_lang] = page_lang
-				data[:page_title] = property['page_title']
+				data[:snapshot_ok_msg] = { :page_id => CGI.escapeHTML(page_id) }
+				data[:page_id_uri] = URI.escape(page_id)
+				data[:page_id] = CGI.escapeHTML(page_id)
+				data[:page_name] = CGI.escapeHTML(page_name)
+				data[:page_lang] = CGI.escapeHTML(page_lang)
+				data[:page_title] = CGI.escapeHTML(property['page_title'])
 			else
-				data[:page_snapshot_auth_fail] = true
-				raise "PageAuthorizationFail"
+				data[:auth_fail] = true
 			end
 		}
 
+		if data[:auth_fail] then
+			raise "PageAuthorizationFail"
+		end
+
 	end
 
 	def index(request_header, request_parameters, data)
@@ -417,9 +440,10 @@ protected
 			page_data = page_data_by_page_name[page_name]
 			unless page_data then
 				page_data = {
-					:page_id => URI.escape(page_name),
+					:page_id_uri => URI.escape(page_name),
+					:page_id => CGI.escapeHTML(page_name),
 					:app_url => @app_url,
-					:page_name => page_name,
+					:page_name => CGI.escapeHTML(page_name),
 					:lang_list => [],
 					'lang_hash' => {},
 				}
@@ -428,26 +452,31 @@ protected
 
 			lang_data = page_data['lang_hash'][lang]
 			unless lang_data then
+				is_frozen = self.site_property['is_frozen']
+				is_frozen = true if (page_property['passwd'].to_s.length > 0)
 				lang_data = {
 					:app_url => @app_url,
-					:page_id => URI.escape("#{page_name}.#{lang}"),
-					:page_lang => lang,
-					:page_title => page_property['page_title'],
+					:page_id_uri => URI.escape("#{page_name}.#{lang}"),
+					:page_id => CGI.escapeHTML("#{page_name}.#{lang}"),
+					:page_lang => CGI.escapeHTML(lang),
+					:page_title => CGI.escapeHTML(page_property['page_title']),
+					:is_frozen => is_frozen, 
 					:count => page_property['count_int'],
 					:version_list => [],
 				}
-				lang_data.update(time2hash(page_property['source_time']))
+				lang_data.update(page_property['source_time'].to_data_hash)
 				page_data['lang_hash'][lang] = lang_data
 				page_data[:lang_list] << lang_data
 			end
 
 			if version then
 				version_data = {
-					:page_id => URI.escape(page_id),
+					:page_id_uri => URI.escape(page_id),
+					:page_id => CGI.escapeHTML(page_id),
 					:app_url => @app_url,
 					:version_int => page_property['version_int'],
 				}
-				version_data.update(time2hash(page_property['source_time']))
+				version_data.update(page_property['source_time'].to_data_hash)
 				lang_data[:version_list] << version_data
 			end
 		}
@@ -471,17 +500,20 @@ protected
 			date_str = page_property['source_time'].strftime("%Y-%m-%d") 
 			date_data = date_hash[date_str]
 			unless date_data then
-				date_data = time2hash(page_property['source_time'])
+				date_data = page_property['source_time'].to_data_hash
 				date_data[:page_list] = []
 				date_hash[date_str] = date_data
 			end
 
 			page_name, lang, version = @page_data_source.parse_page_id(page_id)
-			page_data = time2hash(page_property['source_time'])
+			page_data = page_property['source_time'].to_data_hash
 			page_data[:app_url] = @app_url
-			page_data[:page_id] = page_id
-			page_data[:page_name] = page_name
-			page_data[:page_lang] = lang
+			page_data[:page_id_uri] = URI.escape(page_id)
+			page_data[:page_id] = CGI.escapeHTML(page_id)
+			page_data[:page_name] = CGI.escapeHTML(page_name)
+			page_data[:page_lang] = CGI.escapeHTML(lang)
+			page_data[:is_frozen] = self.site_property['is_frozen']
+			page_data[:is_frozen] = true if (page_property['passwd'].to_s.length > 0)
 			if version then
 				page_data[:version] = { :version_int => page_property['version_int'] }
 			else
@@ -509,7 +541,6 @@ protected
 		default_language = 'en' if default_language.length <= 0
 		languages = request_parameters['languages'].to_s.untaint
 
-
 		@page_data_source.site_property(){ |property|
 			if admin_login_ok?(property, admin_passwd) then
 				# if non passwd or authorized
@@ -545,6 +576,7 @@ protected
 				else
 					property['admin_passwd'] = nil
 				end
+				self.site_property = property
 				data[:admin_passwd_change_ok] = true
 			else
 				data[:admin_auth_fail_msg] = true
@@ -552,13 +584,36 @@ protected
 		}
 	end
 
+	def admin_site_freeze(request_header, request_parameters, data)
+		admin_passwd = request_parameters['admin_passwd'].to_s.untaint
+		site_freeze = request_parameters['site_freeze'].to_s.untaint
+		@page_data_source.site_property(){ |property|
+			if admin_login_ok?(property, admin_passwd) then
+				case site_freeze.downcase 
+				when 'true', 'yes' then
+					property['is_frozen'] = true
+					data[:site_freeze_ok] = true
+				when 'false', 'no' then
+					property['is_frozen'] = false
+					data[:site_thaw_ok] = true
+				end
+				self.site_property = property
+			else
+				data[:admin_auth_fail_msg] = true
+			end
+		}
+	end
+
 	def admin(request_header, request_parameters, data)
 		save_property = request_parameters['save_property'].to_s.untaint
 		change_passwd = request_parameters['change_passwd'].to_s.untaint
+		site_freeze = request_parameters['site_freeze'].to_s.untaint
 		if save_property.length > 0 then
 			admin_save_property(request_header, request_parameters, data)
 		elsif change_passwd.length > 0 then
 			admin_change_passwd(request_header, request_parameters, data)
+		elsif site_freeze.length > 0 then
+			admin_site_freeze(request_header, request_parameters, data)
 		end
 
 		theme_list = Dir.glob("#{self.theme_dir}/*/").sort().collect(){ |theme_dir|
@@ -587,8 +642,45 @@ protected
 		end
 
 		data[:languages] = self.site_property['languages']
+	end
 
+	def delete_page(request_header, request_parameters, data)
+		admin_passwd = request_parameters['admin_passwd'].to_s.untaint
+		@page_data_source.site_property(){ |property|
+			if admin_login_ok?(property, admin_passwd) then
+				page_list = request_parameters['page_list']
+				page_list.each(){ |page_id|
+					name, lang, version = @page_data_source.parse_page_id(page_id)
+					if @page_data_source.exist?(name, lang, version) then
+						@page_data_source.delete_page(name, lang, version)
+					end
+				}
+				data[:delete_ok] = true
+			else
+				data[:admin_auth_fail_msg] = true
+			end
+		}
+	end
 
+	def delete(request_header, request_parameters, data)
+		delete_page = request_parameters['delete_page'].to_s.untaint
+		if delete_page.length > 0 then
+			delete_page(request_header, request_parameters, data)
+		end
+
+		page_list = []
+		@page_data_source.page_list.each(){ |page_info|
+			page_id = page_info[0] 
+			property = page_info[1]
+			page_data = {
+				:app_url => @app_url, 
+				:page_id_uri => URI.escape(page_id), 
+				:page_id => CGI.escapeHTML(page_id), 
+				:page_title => CGI.escapeHTML(property['page_title']),
+			}
+			page_list << page_data
+		}
+		data[:page_list] = page_list
 	end
 
 	def request_command(request_parameters)
@@ -680,6 +772,8 @@ protected
 		result[:default_language] = self.site_property['default_language']
 		result[:language_list] = make_language_list_data(supported_language_list())
 		result[:theme_base_url] = self.theme_base_url
+		result[:site_is_frozen] = self.site_property['is_frozen']
+		result[:site_is_not_frozen] = (not result[:site_is_frozen])
 
 		return result
 	end
@@ -692,16 +786,16 @@ protected
 			property = page[1]
 			n, l, v= @page_data_source.parse_page_id(id)
 			if n == page_name and l == page_lang then
-				version_name = nil
+				is_snapshot = nil
 				if v then
-					version_name = [true, property['version_int'].to_s]
+					is_snapshot = { :version_int => property['version_int'] }
 				end
 				version_list << {
 					:app_url => @app_url,
-					:page_id => id,
-					:recent_name=> (version_name.nil?),
-					:version_name => version_name,
-					:modified_time => time2hash(property['source_time']),
+					:page_id_uri => URI.escape(id),
+					:is_recent=> (is_snapshot.nil?),
+					:is_snapshot => is_snapshot,
+					:modified_time => property['source_time'].to_data_hash,
 				}
 			end
 		}
@@ -719,7 +813,7 @@ protected
 				is_selected = false
 				is_selected = true if page_lang == l 
 				lang_list << {
-					:page_name=> "#{n}.#{l}",
+					:page_id_uri=> URI.escape("#{n}.#{l}"),
 					:is_selected => is_selected,
 					:lang => l,
 				}
@@ -775,16 +869,32 @@ protected
 	end
 
 	def page_login_ok?(page_property, passwd)
+		result = false
+
+		unless self.site_property['is_frozen'] then
+			result = page_owner_login_ok?(page_property, passwd)
+		end
+
+		unless result then
+			if admin_passwd_exist?(self.site_property) then
+				result = admin_login_ok?(self.site_property, passwd)
+			end
+		end
+		
+		return result;
+	end
+
+	def page_owner_login_ok?(page_property, passwd)
+		result = false
 		if page_property['passwd'] and page_property['passwd'].length > 0 then 
-			# page is freezed 
+			# page is frozen 
 			if passwd and page_property['passwd'] == passwd.crypt(page_property['passwd']) then
-				return true;
-			else
-				return false;
+				result = true;
 			end
 		else
-			return true;
+			result =  true;
 		end
+		return result;
 	end
 
 	def admin_passwd_exist?(site_property)
@@ -805,31 +915,34 @@ protected
 		else
 			result = true
 		end
-		result
-	end
-
-	def exception2html(e)
-		result = ''
-		result += '<pre>' 
-		result += CGI.escapeHTML($!.inspect) + "\n"
-		result += CGI.escapeHTML($!.backtrace.join("\n") )
-		result += '</pre>'
-	end
-
-	def time2hash(time)
-		{
-			:yyyy => time.strftime("%Y"),
-			:yy => time.strftime("%y"),
-			:mm => time.strftime("%m"),
-			:dd => time.strftime("%d"),
-			:HH => time.strftime("%H"),
-			:MM => time.strftime("%M"),
-			:SS => time.strftime("%S"),
-			:AA => time.strftime("%A"),
-			:aa => time.strftime("%a"),
-			:BB => time.strftime("%B"),
-			:bb => time.strftime("%b"),
-		}
+		return result
 	end
 end
 
+class Exception
+	def to_html
+		result = self.inspect + "\n" 
+		result += self.to_s + "\n"
+		result += self.backtrace.join("\n") 
+		result = CGI.escapeHTML(result)
+		result = "<pre>#{result}</pre>"
+	end
+end
+
+class Time
+	def to_data_hash
+		{
+			:yyyy => self.strftime("%Y"),
+			:yy => self.strftime("%y"),
+			:mm => self.strftime("%m"),
+			:dd => self.strftime("%d"),
+			:HH => self.strftime("%H"),
+			:MM => self.strftime("%M"),
+			:SS => self.strftime("%S"),
+			:AA => self.strftime("%A"),
+			:aa => self.strftime("%a"),
+			:BB => self.strftime("%B"),
+			:bb => self.strftime("%b"),
+		}
+	end
+end
